#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/un.h>
#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef __CYGWIN__
  #include <cygwin/types.h>
  #include <cygwin/socket.h>
  #define MY_SL_TYPE int
#else
  #include <sys/types.h>
  #include <sys/socket.h>
  #define MY_SL_TYPE unsigned int
#endif

#include <sys/wait.h>

#include "localserver.h"


#define MAX_CONNECTIONS 5


#undef USE_ZOMBIE_CHECK
#ifdef USE_ZOMBIE_CHECK

#define MAX_RUNNING_THREADS 10
static pid_t running_pids[MAX_RUNNING_THREADS];

static int nr_running_threads = 0;

static void InitRunningThreads()
{
  nr_running_threads = 0;
  int i;
  for (i=0;i<MAX_RUNNING_THREADS;i++)
  {
    running_pids[i] = -1;
  }
}

static void AddThreadPid(pid_t newPid)
{
  int i;
  for (i=0;i<MAX_RUNNING_THREADS;i++)
  {
    if (running_pids[i] == -1)
    {
      running_pids[i] = newPid;
      nr_running_threads++;
      break;
    }
  }
}

static void CheckForZombieProcesses()
{
  int i;
  for (i=0;i<MAX_RUNNING_THREADS;i++)
  {
    if (running_pids[i] != -1)
    {
      int status;
      if (wait4(running_pids[i],&status,WNOHANG|WUNTRACED,NULL) == running_pids[i])
      {
        //printf("Found zombie process with id %d\n",running_pids[i]);
        // External application crashed? Bring TTGO to the foreground
        running_pids[i] = -1;
        nr_running_threads--;
      }
    }
  }
}

#else // USE_ZOMBIE_CHECK

static void handle_sigchld(int signo)
{
  int save_errno = errno;
  int status;
  pid_t pid;

  //printf("%s: reaping children\n", __func__);
  for (;;) {
    pid = waitpid(-1, &status, WNOHANG);
    if (pid <= 0)
      break;
    //printf("%s: reaped %d, status %#x\n", __func__, pid, status);
  }
  errno = save_errno;
}

#endif // USE_ZOMBIE_CHECK

static void setsigaction(int num, void (*handler)(int))
{
  struct sigaction sa;
  memset(&sa, 0, sizeof sa);
  sa.sa_handler = handler;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = 0;
  if (sigaction(num, &sa, NULL) == -1)
    err(1, "Unable to set signal handler");
}

int localserver(const char* the_address,void (*requesthandler)(const char* request, int fd))
{
#ifdef USE_ZOMBIE_CHECK
  InitRunningThreads();
#endif // USE_ZOMBIE_CHECK
  int server_sockfd, client_sockfd;
  socklen_t server_len, client_len;
  struct sockaddr_un server_address;
  struct sockaddr_un client_address;
  int result;
  fd_set readfds, testfds;
  int serverbusy;
  serverbusy=1;

  setsigaction(SIGPIPE, SIG_IGN);
#ifndef USE_ZOMBIE_CHECK
  setsigaction(SIGCHLD, handle_sigchld);
#endif // USE_ZOMBIE_CHECK

  unlink(the_address);

  server_sockfd = socket(AF_UNIX, SOCK_STREAM, 0);
  server_address.sun_family = AF_UNIX;
  strcpy(server_address.sun_path, the_address);
  server_len = sizeof(server_address);

  if (bind(server_sockfd, (struct sockaddr *)&server_address, server_len))
  {
    perror("Server Could not bind to the socket\n");
    return 1;
  }
  if (listen(server_sockfd, MAX_CONNECTIONS))
  {
    perror("Server Could not listen to the socket\n");
    return 1;
  }

  FD_ZERO(&readfds);
  FD_SET(server_sockfd, &readfds);

  while(serverbusy)
  {
    int fd;
    int nread;

#ifdef USE_ZOMBIE_CHECK
    CheckForZombieProcesses();
    if (nr_running_threads == MAX_RUNNING_THREADS)
      continue;
#endif // USE_ZOMBIE_CHECK

    //LOGTODO        printf("server waiting\n");
select_retry:
    testfds = readfds;
    result = select(FD_SETSIZE, &testfds, (fd_set *)0, (fd_set *)0, (struct timeval *) 0);

    if(result < 1)
    {
      if (errno == EINTR)
        goto select_retry;
      perror("server5");
      return 1;
    }
    for(fd = 0; fd < FD_SETSIZE; fd++)
    {
      if(FD_ISSET(fd,&testfds))
      {
        if(fd == server_sockfd)
        {
          client_len = sizeof(client_address);
accept_retry:
          client_sockfd = accept(server_sockfd,
                                 (struct sockaddr *)&client_address, (MY_SL_TYPE*)&client_len);
          if (client_sockfd < 0 && errno == EINTR)
            goto accept_retry;
#ifdef __CYGWIN__
          if (client_sockfd != 0xffffffff)
#endif
          {
            FD_SET(client_sockfd, &readfds);
            //LOGTODO                    printf("adding client on fd %d\n", client_sockfd);
          }
        }
        else
        {
          ioctl(fd, FIONREAD, &nread);

          if(nread == 0)
          {
            close(fd);
            FD_CLR(fd, &readfds);
            //LOGTODO                        printf("removing client on fd %d\n", fd);
          }
          else
          {
            pid_t childpid = fork();
            if (childpid == 0)
            {
              //TODO PLEASECHECK optimize!
              char *request = (char *)malloc(nread+1);
              if (request)
              {
                char* buffer = request;
read_retry:
                int result1 = read(fd, buffer, nread);
                if (result1 < 0 && errno == EINTR)
                  goto read_retry;
                buffer[nread]='\0';
                requesthandler(buffer,fd);
                free(request);
              }
              else
              {
                //TODO show an error not enough memory!
              }

              close(fd);
              FD_CLR(fd, &readfds);
              //LOGTODO                        printf("removing client on fd %d\n", fd);
              return 0;
            }
            else
            {
#ifdef USE_ZOMBIE_CHECK
              AddThreadPid(childpid);
#endif // USE_ZOMBIE_CHECK
              close(fd);
              FD_CLR(fd, &readfds);
            }
          }
        }
      }
    }
  }
  return 0;
}
